#include <math.h>
//BMP280
#include <Wire.h>
#include "Adafruit_BME280.h"

//PID
#include <PID_v1.h>

#define I2C_SCL   12
#define I2C_SDA   14

#define PWM_PIN   27
#define PWM_CHANNEL 0     //use first channel of 16 channels
#define PWM_BITS    10    //10 bits for timer
#define PWM_MAX    1023
#define PWM_FREQ    25000 //25kHz
int duty;                 //PWM duty cycle for blower control


#define BME280_ADD1 0x76  //the 2 BME280 share the same I2C bus
#define BME280_ADD2 0x77

#define CPAP_TRIGGER_DELAY 800  //currently blower is On during this constant time (could be changed to become a user's parameter)


void getValues(void);

//BME sensors
Adafruit_BME280 bme1(I2C_SDA, I2C_SCL);
Adafruit_BME280 bme2(I2C_SDA, I2C_SCL);

//BLE
#include <BLEDevice.h>
#include <BLEServer.h>
#include <BLEUtils.h>
#include <BLE2902.h>

//Json
#include <ArduinoJson.h>          //https://github.com/bblanchon/ArduinoJson



long pressureTime ;         //time for last update pressure
int pressureTimeOut = 10 ;  //timeOut for pressure update
float rawPressure1;
float rawPressure2;
float zeroPressure1;        //atmospheric pressure (blower must be OFF)
float zeroPressure2;
float rawTemperature;
float pressure1 = 0;        //blower pressure after substraction of zeroPressure
float pressure2 = 0;
float altitude;
float temperature;
float previousAltitude;
float pressure1MaxFiltered = 0.34;
float pressure1MinFiltered = 0.34;
long pressure1MaxTime;
float diffPressure1 = 0;    //= (pressure1 - zeroPressure1);
float diffPressure2 = 0;    //= (pressure2 - zeroPressure2);
float delta = 0;            //= (pressure1 - zeroPressure1 - pressure2 + zeroPressure2 );
float taux = .03;


int deviceMode = 1;         //0 : BiPAP ; 1: CPAP ; 2 : Spirometer
long cpapTime;
boolean started = false;
float highBreath;           //higher threshold for breathing
float highBreath1;          //higher threshold for breathing
float lowBreath;            //lower threshold for breathing
boolean calibrate = false;
int calibrationStatus = 0;
long calibrationTime;
float breathMin;
float breathMax;
//Spirometer
float tidalVol1;            //tidal volume when blowing
float tidalVol2;            //tidal volume when breathing
long tidalTime1;
long tidalTime2;
boolean started1 = false;   //started blowing
boolean started2 = false;   //started breathing


//PID
double Setpoint;
double Input;
double Output;
double Kp = 20.0;
double Ki = 100.0;
double Kd = 0.0;

PID myPID(&Input, &Output, &Setpoint, Kp, Ki, Kd, DIRECT);
//Input: The variable we're trying to control (double)
//Output: The variable that will be adjusted by the pid (double)
//Setpoint: The value we want to Input to maintain (double)
//Kp, Ki, Kd: Tuning Parameters. these affect how the pid will chage the output. (double>=0)
//Direction: Either DIRECT or REVERSE. determines which direction the output will move when faced with a given error. DIRECT is most common.*/


#define DEBUG
#define W_DEBUG
//#define xDEBUG



//Preferences
#include <Preferences.h>
Preferences preferences;

//BLE declarations
BLECharacteristic *pCharacteristic;
bool deviceConnected = false;
// See the following for generating UUIDs:
// https://www.uuidgenerator.net/

#define SERVICE_UUID        "4fafc201-1fb5-459e-8fcc-c5c9c3319151"
#define CHARACTERISTIC_UUID "beb5483e-36e1-4688-b7f5-ea07361b26a9"

void load_pid_settings (void)
{
  myPID.SetTunings(Kp, Ki, Kd); // send the PID settings to the PID
  myPID.SetOutputLimits(0.0, PWM_MAX);
  myPID.SetSampleTime(10);                    //SampleTime: How often, in milliseconds, the PID will be evaluated. (int>0)
  myPID.SetMode(AUTOMATIC);
}

class MyServerCallbacks: public BLEServerCallbacks {
    void onConnect(BLEServer* pServer) {
      deviceConnected = true;
      Serial.println("client connected");
    };

    void onDisconnect(BLEServer* pServer) {
      deviceConnected = false;
      Serial.println("client disconnected");
    }
};

class MyCallbacks: public BLECharacteristicCallbacks {
    void onWrite(BLECharacteristic *pCharacteristic)
    {
      std::string rxValue = pCharacteristic->getValue();
      String test = "";
      if (rxValue.length() > 0)
      {
        Serial.print("Received : ");
        for (int i = 0; i < rxValue.length(); i++)
        {
          Serial.print(rxValue[i]);
          test = test + rxValue[i];
        }
        Serial.println();
      }
      String Res;

      int i;
      DynamicJsonBuffer jsonBuffer;
      if (test.startsWith("{")) //then it may contain JSON
      {
        String json = test;
        //StaticJsonBuffer<200> jsonBuffer;
        JsonObject& root = jsonBuffer.parseObject(json);

        // Test if parsing succeeds.
        if (!root.success())
        {
          Serial.println("parseObject() failed");
        }
        else
        {
          // Fetch values.
          const char* Chmd = root["cmd"];
          String Cmd(Chmd);

          if (Cmd == "PWM")
          {
            duty = root["value"];
            preferences.putInt("duty", duty);
            //ledcWrite(PWM_CHANNEL, duty);  //set speed to preference value
          }
          else if (Cmd == "PID")         // fetch PID settings
          {
            // Fetch values.
            Kp = root["Kp"];
            Kd = root["Kd"];
            Ki = root["Ki"];
            load_pid_settings();
            Serial.print("PID Kp, Ki, KD ");
            Serial.print(Kp);
            Serial.print("\t");
            Serial.print(Ki);
            Serial.print("\t");
            Serial.println(Kd);
            preferences.putFloat("Kp", Kp);
            preferences.putFloat("Ki", Ki);
            preferences.putFloat("Kd", Kd);
          }
          else  if (Cmd == "Smooth")
          {
            taux = root["value"];
            preferences.putFloat("taux", taux);

          }
          else if (Cmd == "Breath")         // finish decode command  {"cmd":"action,"auto_mode":1,"max_pwm":80,"set_temperature":55.2,"pump":0,"brewery_status":"OK"}
          {
            // Fetch values.
            highBreath = (float)root["highBreath"] / 100.;
            lowBreath = (float)root["lowBreath"] / 100.;
            preferences.putFloat("highBreath", highBreath);
            preferences.putFloat("lowBreath", lowBreath);
          }
          else if (Cmd == "Calibrate")         // finish decode command  {"cmd":"action,"auto_mode":1,"max_pwm":80,"set_temperature":55.2,"pump":0,"brewery_status":"OK"}
          {
            calibrate = true;
            calibrationStatus = 0;
          }
          else if (Cmd == "Mode")         // finish decode command  {"cmd":"action,"auto_mode":1,"max_pwm":80,"set_temperature":55.2,"pump":0,"brewery_status":"OK"}
          {
            // Fetch values.
            deviceMode = root["value"] ;
            preferences.putInt("mode", deviceMode);
            tidalVol1 = 0;
            tidalVol2 = 0;
            Serial.print("mode ");
            Serial.println(deviceMode);
          }
        }
      }
    }
};

//multitasking
TaskHandle_t Task1;


void core0Task1( void * parameter )   //this task will run for ever on core 0
{
  for (;;)                            //forever loop
  {
    delay(100);                       // wait to avoid crash

    //advertise BLE every 100ms
    if ((deviceConnected))
    {
      char message[25];
      int ipressure = (diffPressure1) * 100;
      int ivol1 ;
      if (deviceMode == 2) ivol1 = tidalVol1 * 100;
      else ivol1 = duty / 400.;
      int ivol2 = tidalVol2 * 100;

      sprintf(message, "%06d_%06d_%06d", ipressure, ivol1, ivol2);  //sends values to BA4 APP
      pCharacteristic->setValue(message);
      pCharacteristic->notify();
    }
  }
}


void setup()
{
  Serial.begin(115200);

  //Preferences
  preferences.begin("Spirometer", false);

  //preferences.clear();              // Remove all preferences under the opened namespace
  //preferences.remove("counter");   // remove the counter key only
  duty = preferences.getInt("duty", 0);
  highBreath = preferences.getFloat("highBreath", .4);
  lowBreath = preferences.getFloat("lowBreath", .2);
  deviceMode = preferences.getInt("mode", 1);
  Kp = preferences.getFloat("Kp", 20.);
  Ki = preferences.getFloat("Ki", 100.);
  Kd = preferences.getFloat("Kd", 0.);
  taux = preferences.getFloat("taux", 0.1);
  //preferences.end();  // Close the Preferences


  //BLE
  // Create the BLE Device
  BLEDevice::init("JP eSpirometer");

  // Create the BLE Server
  BLEServer *pServer = BLEDevice::createServer();
  pServer->setCallbacks(new MyServerCallbacks());

  // Create the BLE Service
  BLEService *pService = pServer->createService(SERVICE_UUID);

  // Create a BLE Characteristic
  pCharacteristic = pService->createCharacteristic(
                      CHARACTERISTIC_UUID,
                      BLECharacteristic::PROPERTY_READ   |
                      BLECharacteristic::PROPERTY_WRITE  |
                      BLECharacteristic::PROPERTY_NOTIFY |
                      BLECharacteristic::PROPERTY_INDICATE
                    );

  // https://www.bluetooth.com/specifications/gatt/viewer?attributeXmlFile=org.bluetooth.descriptor.gatt.client_characteristic_configuration.xml
  // Create a BLE Descriptor
  pCharacteristic->addDescriptor(new BLE2902());

  pCharacteristic->setCallbacks(new MyCallbacks());

  // Start the service
  pService->start();

  // Start advertising
  pServer->getAdvertising()->start();
  Serial.println("Waiting a client connection to notify...");
  delay(100);

  //blower
  pinMode(PWM_PIN, OUTPUT);
  digitalWrite(PWM_PIN, LOW);
  ledcSetup(PWM_CHANNEL, PWM_FREQ, PWM_BITS); //configure PWM
  ledcAttachPin(PWM_PIN, PWM_CHANNEL); // attach the channel to the GPIO to be controlled
  ledcWrite(PWM_CHANNEL, 0);  //set speed to stop value

  //BME280
  bool status;

  status = bme1.begin(BME280_ADD1); //need to modifiy BME280.cpp the library to use this lib with BMP280...
  //  // check if sensor, i.e. the chip ID is correct
  // if ((read8(BME280_REGISTER_CHIPID) != 0x60)&&(read8(BME280_REGISTER_CHIPID) != 0x58))
  //    return false;
  if (!status) {
    Serial.println("Could not find a valid BME280 sensor, check wiring!");
    while (1);
  }
  status = bme2.begin(BME280_ADD2); //need to modifiy BME280.cpp the library to use this lib with BMP280...
  //  // check if sensor, i.e. the chip ID is correct
  // if ((read8(BME280_REGISTER_CHIPID) != 0x60)&&(read8(BME280_REGISTER_CHIPID) != 0x58))
  //    return false;
  if (!status) {
    Serial.println("Could not find a valid BME280 sensor, check wiring!");
    while (1);
  }

  /*modify here to set default values in BME280.h library
    void setSampling(sensor_mode mode              = MODE_NORMAL,
         sensor_sampling tempSampling  = SAMPLING_X2,
         sensor_sampling pressSampling = SAMPLING_X16,
         sensor_sampling humSampling   = SAMPLING_X16,
         sensor_filter filter          = FILTER_X16,
         standby_duration duration     = STANDBY_MS_0_5
         );*/

  getZeroValues(); //calibrate zero values of pressure sensors


  //PID
  load_pid_settings ();


  //multitask
  xTaskCreatePinnedToCore(
    core0Task1,               /* Task function. */
    "Task_1",                 /* name of task. */
    5000,                     /* Stack size of task */
    NULL,                     /* parameter of the task */
    1,                        /* priority of the task */
    &Task1,                   /* Task handle to keep track of created task */
    0);                       /* Core */

  //CPAP
  cpapTime = millis();
  pressure1MaxTime = millis();
}


//Main loop repeated for ever
void loop()
{
  if ((millis() - pressureTime) > pressureTimeOut)
  {
    pressureTime = millis();

    getValues();  //read sensors and substract atmospheric pressure and filter

    //    if (deviceMode > 0) //avoid auto calibration in biPAP
    //    {
    //      if (abs(diffPressure1) < 0.02) zeroPressure1 = 0.8 * zeroPressure1 + 0.2 * pressure1; //compute filtered zero values
    //      if (abs(diffPressure2) < 0.02 ) zeroPressure2 = 0.8 * zeroPressure2 + 0.2 * pressure2;
    //    }
    //calibration
    if (calibrate == true)
    {
      switch (calibrationStatus)
      {
        case 0:                     //stop blower and calibrate zero
          duty = 0;
          ledcWrite(PWM_CHANNEL, 0);  //stop the blower
          if ((millis() - calibrationTime) > 20000)
          {
            getZeroValues();
            calibrationStatus++;
            calibrationTime = millis();
            Setpoint = preferences.getInt("duty") / -3.2; //restart the blower
          }
          break;
        default:
          calibrate = false;
          calibrationStatus = 0;
          break;
      }
    }
    else      //normal operation mode
    {
      if (diffPressure1 > breathMax) breathMax = diffPressure1; //compute filtered max pressure
      if (diffPressure1 < breathMin) breathMin = diffPressure1; //compute filtered min pressure

      if (breathMax > 0.1)
      {
        pressure1MaxFiltered = 0.2 * breathMax + 0.8 * pressure1MaxFiltered;
        breathMax = -10000.;
        pressure1MaxTime = millis();
        highBreath = 0.65 * pressure1MaxFiltered; //thresholds used to trigger blower according to breathing rate
        lowBreath = 0.3 * pressure1MaxFiltered;
      }

      // PID run

      switch (deviceMode)
      {
        case 0: //BiPAP
          Input = 1024 - sqrt(fmax((-delta), 0.)) * 600; //control loop on flow value (delta value between two sensors is proportiennal to square of velocity (ie square of flow rate)
          if ((diffPressure1 > highBreath))
          {
            if ((millis() - cpapTime) >  CPAP_TRIGGER_DELAY)
            {
              started = false;
              Setpoint = 0;
            }
          }
          else
          {
            if (((diffPressure1 < lowBreath)) && (started == false))
            {
              started = true;
              Setpoint = duty; /// 3.2;
              cpapTime = millis();
            }
          }
          break;
        case 1: //CPAP
          Input = (diffPressure1 * 600);          //control loop on pressure value  (venturi not needed, only one single sensor)
          Setpoint = duty ;
          break;
        case 2: //Spirometer
          //blowing
          if ((diffPressure1 < -.2))  //inspire
          {
            if ((millis() - tidalTime1) >  CPAP_TRIGGER_DELAY)
            {
              started1 = false;
              tidalVol1 = 0;
            }
          }
          else
          {
            if (((diffPressure1 > 0.1)) && (started1 == false)) //blowing start and integrate flow
            {
              started1 = true;
              tidalTime1 = millis();
            }
            if ((started1 == true) && (delta < -0.02 ))
            {
              tidalVol1 += sqrt(-delta);
            }
          }
          //breathing
          if ((diffPressure1 > 0.2))  //blowing
          {
            if ((millis() - tidalTime2) >  CPAP_TRIGGER_DELAY)
            {
              started2 = false;
              tidalVol2 = 0;
            }
          }
          else
          {
            if (((diffPressure1 < -0.2)) && (started2 == false)) //breathing start and integrate flow
            {
              started2 = true;
              tidalTime2 = millis();
              tidalVol2 = 0;
            }
            if ((started2 == true) && (delta > 0.02))
            {
              tidalVol2 += sqrt(delta);
            }
          }
          break;
        default:
          // statements
          break;
      }
    }
  }
  // blower
  if (deviceMode < 2)   //activate blower excepted in spirometer mode
  {
    myPID.Compute();
    if (Setpoint > 1) ledcWrite(PWM_CHANNEL, Output);
    else ledcWrite(PWM_CHANNEL, 0);                   //set speed to PID Output value
  }


#ifdef DEBUG
  Serial.print(diffPressure1 * 600);
  Serial.print("\t");
  //  Serial.print(diffPressure2);
  //  Serial.print("\t");
  Serial.print(Setpoint) ;
  Serial.print("\t");
  //  Serial.print(highBreath );
  //  Serial.print("\t");
  //  Serial.print(lowBreath );
  //  Serial.print("\t");
  //  Serial.print(tidalVol1 );
  //  Serial.print("\t");
  //  Serial.print(tidalVol2 );
  //  Serial.print("\t");
  Serial.print(Input );
  Serial.print("\t");
  Serial.print(Output);
  Serial.print("\t");
  Serial.println(" hPa");
#endif
}


void getValues() //read BME280 values
{
  rawPressure1 = (bme1.readPressure()) / 100.0F;  //result in hPa
  rawPressure2 = (bme2.readPressure()) / 100.0F;
  //  rawTemperature = bme2.readTemperature();
  //pressure differences and filter
  diffPressure1 = ((1 - taux) * diffPressure1) + taux * (rawPressure1 - zeroPressure1);
  diffPressure2 = (1 - taux) * diffPressure2 + taux * (rawPressure2 - zeroPressure2);
  delta = diffPressure1 - diffPressure2;
}


void getZeroValues(void)  //compute zero values as an average during 100 readings
{
  int i = 0;
  zeroPressure1 = 0;
  zeroPressure2 = 0;
  for (i = 0; i < 100; i++)
  {
    getValues();
    delay(6); //sensor has an internal delay of 5ms between samples (see BME280.h)
    zeroPressure1 += rawPressure1;
    zeroPressure2 += rawPressure2;
  }
  zeroPressure1 = zeroPressure1 / i;  //calibration of atmospheric pressure
  zeroPressure2 = zeroPressure2 / i;
}
